##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""

from p01.jsonrpc import interfaces


class IJSONRPCLayer(interfaces.IJSONRPCRequest):
    """Layer for JSONRPC requests.

    IMPORTANT:
    If you like to use ZPT templates and call adatapers in MethodPublisher 
    views, you need to make sure that your layer provides the relevant
    adapters. This means, you will probably use a layer inherited from 
    IBrowserRequest and IJSONRPCRequest in your skin.

    Or you can register your icons etc. for the IJSONRequest layer etc. if you
    need to return content produced by ZPT tempaltes. Also note that icons
    or other resources are only available in IBrowserRequest or in your custom
    layer.
    """



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""

import zope.component
from zope.publisher.interfaces import NotFound
from p01.jsonrpc.interfaces import IMethodPublisher


class JSONRPCTraversablePage(object):
    """Mixin for JSON-RPC traversable views, pages, forms etc.
    
    Make sure if you inherit from this class that no other super class
    will override the publishTraverse method.
    """

    def publishTraverse(self, request, name):
        view = zope.component.queryMultiAdapter((self, request), name=name)
        if view is None or not IMethodPublisher.providedBy(view):
            raise NotFound(self, name, request)
        return view



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""

import zope.i18n
import zope.i18nmessageid

_ = zope.i18nmessageid.MessageFactory('p01')

from p01.jsonrpc import interfaces

# The error codes used since JSON-RPC 2.0
# See: http://groups.google.com/group/json-rpc/web/json-rpc-1-2-proposal
#
# code   message           Meaning 
# --------------------------------
# -32700 Parse error.          Invalid JSON. An error occurred on the server 
#                              while parsing the JSON text.
# -32600 Invalid Request.      The received JSON not a valid JSON-RPC Request.
# -32601 Method not found.     The requested remote-procedure does not exist, 
#                              is not available.
# -32602 Invalid params.       Invalid method parameters.
# -32603 Internal error.       Internal JSON-RPC error.
# -32099..-32000 Server error. Reserved for implementation-defined server-errors.


class JSONRPCErrorView(object):
    """Generic JSON-RPC error view.
    
    This base class is used for error views which are used for error handling
    by ZopePublication.
    """
    
    zope.interface.implements(interfaces.IJSONRPCErrorView)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        """Must return itself.
        
        This allows us to use the error view in setResult if ZopePublication
        is adapting an error view to error and request and calls them.
        """
        return self


class ParseErrorView(JSONRPCErrorView):
    """Knows the error data for parse errors."""

    code = -32700
    message = u'Parse error'

    @property
    def data(self):
        errMsg = _('Parse error')
        return {'i18nMessage':zope.i18n.translate(errMsg, context=self.request)}


class InvalidRequestErrorView(JSONRPCErrorView):
    """Knows the error data for invalid request errors."""

    code = -32600
    message = u'Invalid Request'

    @property
    def data(self):
        errMsg = _('Invalid Request')
        return {'i18nMessage':zope.i18n.translate(errMsg, context=self.request)}


class MethodNotFoundView(JSONRPCErrorView):
    """Knows the error data for NotFound errors."""

    code = -32601
    message = u'Method not found'

    @property
    def data(self):
        errMsg = _('Method not found')
        return {'i18nMessage':zope.i18n.translate(errMsg, context=self.request)}


class InvalidParamsErrorView(JSONRPCErrorView):
    """Knows the error data for invalid params errors."""

    code = -32602
    message = u'Invalid params'

    @property
    def data(self):
        errMsg = _('Invalid params')
        return {'i18nMessage':zope.i18n.translate(errMsg, context=self.request)}


class InternalErrorView(JSONRPCErrorView):
    """Knows the error data for invalid params errors."""

    code = -32603
    message = u'Internal error'

    @property
    def data(self):
        errMsg = _('Internal error')
        return {'i18nMessage':zope.i18n.translate(errMsg, context=self.request)}



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import zope.interface
import zope.component
from zope.app.publication.http import BaseHTTPPublication
from zope.app.publication.interfaces import IRequestPublicationFactory
from zope.app.publication.interfaces import IPublicationRequestFactory

from p01.jsonrpc import interfaces
from p01.jsonrpc.publisher import JSONRPCRequest


class JSONRPCPublication(BaseHTTPPublication):
    """JSON RPC Publication."""

    zope.interface.implements(interfaces.IJSONRPCPublication)


class JSONRPCFactory(object):
    zope.interface.implements(IRequestPublicationFactory)
    
    def canHandle(self, environment):
        return True
        
    def __call__(self):
        request_class = zope.component.queryUtility(
            interfaces.IJSONRPCRequestFactory, default=JSONRPCRequest)
        return request_class, JSONRPCPublication



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import unittest

import zope.interface
import zope.component
from zope.component.testing import PlacelessSetup

from zope.app.publication.httpfactory import HTTPPublicationRequestFactory

from p01.jsonrpc.publication import JSONRPCFactory
from p01.jsonrpc.publication import JSONRPCPublication
from p01.jsonrpc import interfaces


class DummyRequestFactory(object):
    def __call__(self, input_stream, env):
        self.input_stream = input_stream
        self.env = env
        return self

    def setPublication(self, pub):
        self.pub = pub


class Test(PlacelessSetup, unittest.TestCase):

    def setUp(self):
        super(Test, self).setUp()
        self.__factory = HTTPPublicationRequestFactory(None)
        self.__env =  {
            'SERVER_URL':         'http://127.0.0.1',
            'HTTP_HOST':          '127.0.0.1',
            'CONTENT_LENGTH':     '0',
            'GATEWAY_INTERFACE':  'TestFooInterface/1.0'
            }

    def test_jsonrpcfactory(self):
        jsonrpcrequestfactory = DummyRequestFactory()
        zope.interface.directlyProvides(
            jsonrpcrequestfactory, interfaces.IJSONRPCRequestFactory)
        zope.component.provideUtility(jsonrpcrequestfactory)
        env = self.__env
        factory = JSONRPCFactory()
        self.assertEqual(factory.canHandle(env), True)
        request, publication = factory()
        self.assertEqual(isinstance(request, DummyRequestFactory), True)
        self.assertEqual(publication, JSONRPCPublication)


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(Test),
        ))


if __name__=='__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import unittest
from StringIO import StringIO

import zope.component
from zope.publisher.base import DefaultPublication
from zope.publisher.http import HTTPCharsets

from p01.jsonrpc.publisher import JSONRPCRequest


class Publication(DefaultPublication):

    require_docstrings = 0


class TestJSONRPCRequest(JSONRPCRequest, HTTPCharsets):
    """Make sure that our request also implements IHTTPCharsets, so that we do
    not need to register any adapters."""

    def __init__(self, *args, **kw):
        self.request = self
        JSONRPCRequest.__init__(self, *args, **kw)


class PositionalTestCall:
    def __init__(self):
        self.body = '{"jsonrpc":"2.0","id":"httpReq","method":"positional","params":["aaa","bbb"]}'
        self.headers = []


class NamedTestCall:
    def __init__(self):
        self.body = '{"jsonrpc":"2.0","id":"httpReq","method":"named","params":{"kw1":"aaa","kw2":"bbb"}}'
        self.headers = []


class JSONRPCTests(unittest.TestCase):
    """The only thing different to HTTP is the input processing; so there
       is no need to redo all the HTTP tests again.
    """

    _testEnv =  {
        'PATH_INFO':          '/folder/item2/view/',
        'QUERY_STRING':       '',
        'SERVER_URL':         'http://foobar.com',
        'HTTP_HOST':          'foobar.com',
        'CONTENT_LENGTH':     '0',
        'REQUEST_METHOD':     'POST',
        'HTTP_AUTHORIZATION': 'Should be in accessible',
        'GATEWAY_INTERFACE':  'TestFooInterface/1.0',
        'HTTP_OFF_THE_WALL':  "Spam 'n eggs",
        'HTTP_ACCEPT_CHARSET': 'ISO-8859-1, UTF-8;q=0.66, UTF-16;q=0.33',
    }

    def setUp(self):
        super(JSONRPCTests, self).setUp()

        class AppRoot(object):
            pass

        class Folder(object):
            pass

        class Item(object):

            def __call__(self, a, b):
                return "%s, %s" % (a, b)

            def doit(self, a, b):
                return 'do something %s %s' % (a, b)

        class View(object):

            def __init__(self, context, request):
                self.context = context
                self.request = request

            def positional(self, a, b):
                return "Parameter: a; type: %s; value: %s" %(
                    type(a).__name__, a)

            def named(self, kw1=None, kw2=None):
                kw1 = self.request.get('kw1', kw1)
                kw2 = self.request.get('kw2', kw2)
                return "Keyword: kw1; type: %s; value: %s] " \
                       "Keyword: kw2; type: %s; value: %s]" %(
                    type(kw1).__name__, kw1, type(kw2).__name__, kw2)

        class Item2(object):

            request = None

            @property
            def view(self):
                return View(self, self.request)

        self.app = AppRoot()
        self.app.folder = Folder()
        self.app.folder.item = Item()
        self.app.folder.item2 = Item2()
        # zope.component.provideUtility(JSONReader(), IJSONReader)

    def _createRequest(self, extra_env={}, body=""):
        env = self._testEnv.copy()
        env.update(extra_env)
        if len(body.body):
            env['CONTENT_LENGTH'] = str(len(body.body))
        publication = Publication(self.app)
        instream = StringIO(body.body)
        request = TestJSONRPCRequest(instream, env)
        request.setPublication(publication)
        return request


    def testProcessInput(self):
        req = self._createRequest({}, PositionalTestCall())
        self.app.folder.item2.request = req
        req.processInputs()
        self.assertEqual(req._args, (u'aaa', u'bbb'))
        self.assertEqual(tuple(req._path_suffix), ('positional',))

    def testPositional(self):
        req = self._createRequest({}, PositionalTestCall())
        self.app.folder.item2.request = req
        req.processInputs()
        action = req.traverse(self.app)
        self.assertEqual(action(*req._args),
                             'Parameter: a; type: unicode; value: aaa')

    def testKeyword(self):
        req = self._createRequest({}, NamedTestCall())
        self.app.folder.item2.request = req
        req.processInputs()
        action = req.traverse(self.app)
        self.assertEqual(action(*req._args, **req.form),
            u'Keyword: kw1; type: unicode; value: aaa] Keyword: kw2; type: unicode; value: bbb]')

    def testJSONRPCMode(self):
        req = self._createRequest({}, PositionalTestCall())
        self.app.folder.item2.request = req
        req.processInputs()
        self.assertEqual(req['JSONRPC_MODE'],True)


def test_suite():
    loader = unittest.TestLoader()
    return loader.loadTestsFromTestCase(JSONRPCTests)


if __name__=='__main__':
    unittest.TextTestRunner().run(test_suite())



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import re
import sys
import logging

import zope.interface
import zope.component
from zope.exceptions.exceptionformatter import format_exception
from zope.location.location import Location
from zope.i18n.interfaces import IUserPreferredCharsets
from zope.publisher.http import HTTPRequest
from zope.publisher.http import HTTPResponse
from zope.publisher.http import getCharsetUsingRequest
from zope.publisher.http import DirectResult
from zope.security.proxy import isinstance

import p01.json.api

import p01.jsonrpc.exception
from p01.jsonrpc import interfaces


JSON_RPC_VERSION = '2.0'

DEBUG = logging.DEBUG
logger = logging.getLogger()


def intsort(item):
    return int(item[0])


class MethodPublisher(Location):
    """Base class for JSON-RPC views that publish methods
       like zope.app.publisher.xmlrpc.MethodPublisher
    """
    zope.interface.implements(interfaces.IMethodPublisher)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __getParent(self):
        return hasattr(self, '_parent') and self._parent or self.context

    def __setParent(self, parent):
        self._parent = parent

    __parent__ = property(__getParent, __setParent)


class MethodTraverser(object):
    zope.interface.implements(interfaces.IJSONRPCPublisher)

    __used_for__ = interfaces.IMethodPublisher

    def __init__(self, context, request):
        self.context = context

    def publishTraverse(self, request, name):
        return getattr(self.context, name)


class JSONRPCRequest(HTTPRequest):
    """JSON-RPC request implementation based on IHTTPRequest.

    This implementation supports the following JSON-RPC Specification versions:

    - 1.0
    - 1.1
    - 2.0

    Version 1.0 and 1.1 offers params as a list. This params get converted to
    positional arguments if calling the JSON-RPC function.

    The version 2.0 offers support for named key/value params. The important
    thing to know is that this implementation will convert named params kwargs
    to form paramters. This means the method doesn't get any key word argument.
    The reason why I was choosing is the existing publisher implementation and
    it's debugger integration. If someone likes to integrate **kwargs support,
    take a look at the publisher.publish method and it's mapply function which
    get wrapped by the Debugger class. I hope that's fine for now and I
    recommend to avoid kwargs for JSON-RPC methods ;-)

    The p01.jsonrpcclient JavaScript method JSONRPCProxy converts a
    typeof object as arguments[0] to named key/value pair arguments.

    """

    _jsonId = 'jsonrpc'
    jsonVersion = JSON_RPC_VERSION
    jsonId = None

    zope.interface.implements(interfaces.IJSONRPCRequest,
                              interfaces.IJSONRPCApplicationRequest)

    def __init__(self, body_instream, environ, response=None):
        self.form = {}
        self._args = ()
        self.charsets = None
        super(JSONRPCRequest, self).__init__(body_instream, environ, response)

    def _createResponse(self):
        """return a response"""
        return JSONRPCResponse()

    def _decode(self, text):
        """Try to decode the text using one of the available charsets."""
        if self.charsets is None:
            envadapter = IUserPreferredCharsets(self)
            self.charsets = envadapter.getPreferredCharsets() or ['utf-8']
        for charset in self.charsets:
            try:
                text = unicode(text, charset)
                break
            except UnicodeError:
                pass
        return text

    _typeFormat = re.compile('([a-zA-Z][a-zA-Z0-9_]+|\\.[xy])$')

    def processInputs(self):
        """take the converted request and make useful args of it."""
        # from z3c.json.interfaces import IJSONReader
        # json = zope.component.getUtility(IJSONReader)
        stream = self._body_instream
        input = []
        incoming = stream.read(1000)
        while incoming:
            input.append(incoming)
            incoming = stream.read(1000)
        input = ''.join(input)
        # ensure unicode
        if not isinstance(input, unicode):
            input = self._decode(input)
        try:
            #data = json.read(input)
            data = p01.json.api.jsonReader(input)
        except:
            # catch any error since we don't know which library is used as
            # parser
            raise p01.jsonrpc.exception.ParseError
        # get the params
        params = data.get('params', [])
        if self.jsonId is None:
            self.jsonId = data.get('id', self._jsonId)

        # get the json version. The version 1.0 offers no version argument.
        # The version 1.1 offers a version key and since version 2.0 the
        # version is given with the ``jsonrpc`` key. Let's try to find the
        # version for our request.
        self.jsonVersion = data.get('version', self.jsonVersion)
        self.jsonVersion = data.get('jsonrpc', self.jsonVersion)
        if self.jsonVersion in ['1.0', '1.1', '2.0']:
            # json-rpc 1.0 and 1.1
            if isinstance(params, list):
                args = params
                # version 1.0 and 1.1 uses a list of arguments
                for arg in args:
                    if isinstance(arg, dict):
                        # set every dict key value as form items and support at
                        # least ``:list`` and ``:tuple`` input field name postifx
                        # conversion.
                        for key, d in arg.items():
                            key = str(key)
                            pos = key.rfind(":")
                            if pos > 0:
                                match = self._typeFormat.match(key, pos + 1)
                                if match is not None:
                                    key, type_name = key[:pos], key[pos + 1:]
                                    if type_name == 'list' and not isinstance(d, list):
                                        d = [d]
                                    if type_name == 'tuple' and not isinstance(d, tuple):
                                        d = tuple(d)
                            self.form[key] = d
            elif isinstance(params, dict):
                # process the key/value pair params. This arguments get stored
                # in the request.form argument and we skip it from method calls.
                # This means this library will not support key word arguments
                # for method calls. It will instead store them in the form.
                # This has two reasons.
                # 1. Zope doesn't support kwargs in the publication
                #    implementation. It only supports positional arguments
                # 2. The JSON-RPC specification doesn't allow to use positional
                #     and keyword arguments on one method call
                # 3. Python doesn't allow to convert kwargs to positional
                #    arguments because a dict doesn't provide an order
                # This means you should avoid to call a method with kwargs.
                # just use positional arguments if possible. Or get them from
                # directly from the request or request.form argument in your
                # code. Let me know if this is a real problem for you and you
                # like to implement a different kwarg handling. We have some
                # ideas for add support for this.
                args = params
                # set every dict key value as form items and support at
                # least ``:list`` and ``:tuple`` input field name postifx
                # conversion.
                for key, d in args.items():
                    key = str(key)
                    pos = key.rfind(":")
                    if pos > 0:
                        match = self._typeFormat.match(key, pos + 1)
                        if match is not None:
                            key, type_name = key[:pos], key[pos + 1:]
                            if type_name == 'list' and not isinstance(d, list):
                                d = [d]
                            if type_name == 'tuple' and not isinstance(d, tuple):
                                d = tuple(d)
                    self.form[key] = d
                args = []
            elif params is None:
                args = []
        else:
            raise TypeError(
                'Unsupported JSON-RPC version (%s)' % self.jsonVersion)
        self._args = tuple(args)
        # make environment, cookies, etc., available to request.get()
        super(JSONRPCRequest, self).processInputs()
        self._environ['JSONRPC_MODE'] = True

        # split here on '.' for get path suffix steps
        functionstr = data['method']
        function = functionstr.split('.')
        if function:
            # translate '.' to '/' in function to represent object traversal.
            self.setPathSuffix(function)

    def traverse(self, object):
        return super(JSONRPCRequest, self).traverse(object)

    def keys(self):
        """See Interface.Common.Mapping.IEnumerableMapping."""
        d = {}
        d.update(self._environ)
        d.update(self._cookies)
        d.update(self.form)
        return d.keys()

    def get(self, key, default=None):
        """See Interface.Common.Mapping.IReadMapping."""
        marker = object()
        result = self.form.get(key, marker)
        if result is not marker:
            return result
        return super(JSONRPCRequest, self).get(key, default)

    def __getitem__(self, key):
        return self.get(key)


class JSONRPCResponse(HTTPResponse):
    """JSON-RPC Response"""

    def setResult(self, result):
        """The result dict contains the following key value pairs

        The version 1.0 and 1.1 provides a response dict with the following
        arguments:

        id -- json request id
        result -- result or null on error
        error -- error or null if result is Ok

        The version 2.0 provides a response dict with the following named
        paramters:

        jsonrpc -- jsonrpc version 2.0 or higher in future versions
        id -- json request id
        result -- result if no error is raised
        error -- error if any given

        """
        jsonId = self._request.jsonId
        jsonVersion = self._request.jsonVersion

        if interfaces.IJSONRPCErrorView.providedBy(result):
            if self._request.jsonVersion == "1.0":
                wrapper = {'result': None,
                           'error': result.message,
                           'id': self._request.jsonId}
            elif self._request.jsonVersion == "1.1":
                wrapper = {'version': self._request.jsonVersion,
                           'error': result.message,
                           'id': self._request.jsonId}
            else:
                wrapper = {'jsonrpc': self._request.jsonVersion,
                           'error': {'code': result.code,
                                     'message': result.message,
                                     'data': result.data},
                           'id': self._request.jsonId}

            try:
            # from z3c.json.interfaces import IJSONWriter
                # json = zope.component.getUtility(IJSONWriter)
                # result = json.write(wrapper)
                result = p01.json.api.jsonWriter(wrapper)
                body = self._prepareResult(result)
                super(JSONRPCResponse, self).setResult(DirectResult((body,)))
                logger.log(DEBUG, "Exception: %s" % result)
                # error response is not really an error, it's valid response
                self.setStatus(200)
            except:
                # Catch all exceptions at this point
                self.handleException(sys.exc_info())
                return
        else:
            # from z3c.json.converter import premarshal
            # result = premarshal(result)
            if jsonVersion == "1.0":
                wrapper = {'result': result, 'error': None, 'id': jsonId}
            elif jsonVersion == "1.1":
                wrapper = {
                    'version': jsonVersion, 'result': result, 'id': jsonId}
            else:
                wrapper = {
                    'jsonrpc': jsonVersion, 'result': result, 'id': jsonId}
            # from z3c.json.interfaces import IJSONWriter
            # json = zope.component.getUtility(IJSONWriter)
            # result = json.write(wrapper)
            encoding = getCharsetUsingRequest(self._request)
            result = p01.json.api.jsonWriter(wrapper)
            body = self._prepareResult(result)
            super(JSONRPCResponse, self).setResult(DirectResult((body,)))
            logger.log(DEBUG, "%s" % result)

    def _prepareResult(self, result):
        # we've asked json to return unicode; result should be unicode
        encoding = getCharsetUsingRequest(self._request) or 'utf-8'
        enc = encoding.lower()
        if not enc in interfaces.JSON_CHARSETS:
            encoding = 'utf-8'
        # encode outgoing boundary.
        if isinstance(result, unicode):
            body = result.encode(encoding)
            charset = encoding
        else:
            # something's wrong. JSON did not return unicode.
            raise TypeError("JSON did not return unicode (%s)" % type(result))

        # set content type
        self.setHeader('content-type', "application/x-javascript;charset=%s"
                       % charset)
        return body

    def handleException(self, exc_info):
        # only legacy Exception where we didn't define a view for get handled
        # by this method. All exceptions where we have a view registered for
        # get handled by the setResult method based on the given
        # IJSONRPCErrorView
        logger.log(logging.ERROR, "".join(format_exception(
            exc_info[0], exc_info[1], exc_info[2], with_filenames=True)))

        t, value = exc_info[:2]
        s = '%s: %s' % (getattr(t, '__name__', t), value)
        if self._request.jsonVersion == "1.0":
            wrapper = {'result': None,
                       'error': s,
                       'id': self._request.jsonId}
        elif self._request.jsonVersion == "1.1":
            wrapper = {'version': self._request.jsonVersion,
                       'error': s,
                       'id': self._request.jsonId}
        else:
            # this only happens if error handling was running into en error or
            # if we didn't define an IJSONRPCErrorView for a given error
            wrapper = {'jsonrpc': self._request.jsonVersion,
                       'error': {'code': -32603,
                                 'message': 'Internal error',
                                 'data': s},
                       'id': self._request.jsonId}

        # from z3c.json.interfaces import IJSONWriter
        # json = zope.component.getUtility(IJSONWriter)
        # result = json.write(wrapper)
        result = p01.json.api.jsonWriter(wrapper)
        body = self._prepareResult(result)
        super(JSONRPCResponse, self).setResult(DirectResult((body,)))
        logger.log(DEBUG, "Exception: %s" % result)
        self.setStatus(200)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import base64
import copy
import re
import rfc822
import sys
import json
import doctest
import persistent
from cStringIO import StringIO

PY3 = sys.version_info[0] == 3
if PY3: # pragma: no coverr
    from urllib import parse as urlparse
    from http.cookies import SimpleCookie
    def unquote(string):
        if not string:
            return b''
        res = string.split(b'%')
        if len(res) != 1:
            string = res[0]
            for item in res[1:]:
                try:
                    string += bytes([int(item[:2], 16)]) + item[2:]
                except ValueError:
                    string += b'%' + item
        return string
    def url_unquote(s):
        return unquote(s.encode('ascii')).decode('latin-1')
else:
    import urlparse
    from Cookie import SimpleCookie
    from urllib import unquote as url_unquote

import transaction

import zope.interface
import zope.component
import zope.testing.cleanup
import zope.container.interfaces
from zope.component import hooks

import p01.publisher.interfaces
import p01.publisher.application
import p01.publisher.testing

import p01.json.exceptions
import p01.json.proxy
import p01.json.transport

from p01.jsonrpc import layer
from p01.jsonrpc.publication import JSONRPCPublication
from p01.jsonrpc.publisher import JSONRPCRequest
from p01.jsonrpc.publisher import MethodPublisher
from p01.jsonrpc.publisher import JSON_RPC_VERSION

SCHEME_RE = re.compile(r'^[a-z]+:', re.I)


###############################################################################
#
# test application
#
###############################################################################

class IJSONRPCTestApplication(zope.container.interfaces.IContainer,
        p01.publisher.interfaces.IApplication):
    """Test application offering container"""

    def get(key, default=None):
        """Returns item or defualt"""

    def __getitem__(key):
        """Returns item or defualt"""


class JSONRPCTestApplication(p01.publisher.application.Application):
    """Test application"""

    zope.interface.implements(IJSONRPCTestApplication)

    def __init__(self):
        super(JSONRPCTestApplication, self).__init__()
        self._data = {}

    def __setitem__(self, key, obj):
        self._data[key] = obj
        obj.__parent__ = self
        obj.__name__ = key

    def __getitem__(self, key):
        return self._data[key]

    def get(self, key, default=None):
        try:
            return self._data[key]
        except KeyError as e:
            return default

    def __delitem__(self, key):
        del self._data[key]

    def __repr__(self):
        return '<%s>' % (self.__class__.__name__)


def setUpTestApplication():
    """Setup test application used as root and for test traversing"""
    app = JSONRPCTestApplication()
    zope.component.provideUtility(app, p01.publisher.interfaces.IApplication,
        '')
    return app


def getRootFolder():
    return zope.component.getUtility(p01.publisher.interfaces.IApplication)


headerre = re.compile(r'(\S+): (.+)$')
def split_header(header):
    return headerre.match(header).group(1, 2)


basicre = re.compile('Basic (.+)?:(.+)?$')
def auth_header(header):
    match = basicre.match(header)
    if match:
        u, p = match.group(1, 2)
        if u is None:
            u = ''
        if p is None:
            p = ''
        auth = base64.encodestring('%s:%s' % (u, p))
        return 'Basic %s' % auth[:-1]
    return header


###############################################################################
#
# JSONRPC Test proxy
#
###############################################################################

class ResponseWrapper(object):
    """A wrapper that adds several introspective methods to a response."""

    def __init__(self, response, path, omit=()):
        self._response = response
        self._path = path
        self.omit = omit
        self._body = None

    def getOutput(self):
        """Returns the full HTTP output (headers + body)"""
        body = self.getBody()
        omit = self.omit
        headers = [x
                   for x in self._response.getHeaders()
                   if x[0].lower() not in omit]
        headers.sort()
        headers = '\n'.join([("%s: %s" % (n, v)) for (n, v) in headers])
        statusline = '%s %s' % (self._response._request['SERVER_PROTOCOL'],
                                self._response.getStatusString())
        if body:
            return '%s\n%s\n\n%s' %(statusline, headers, body)
        else:
            return '%s\n%s\n' % (statusline, headers)

    def getBody(self):
        """Returns the response body"""
        if self._body is None:
            self._body = ''.join(self._response.consumeBody())

        return self._body

    def getPath(self):
        """Returns the path of the request"""
        return self._path

    def __getattr__(self, attr):
        return getattr(self._response, attr)

    __str__ = getOutput


class JSONRPCTestTransport(p01.json.transport.Transport):
    """Test transport using wsgi application and it's publisher.

    It can be used like a normal transport, including support for basic
    authentication.
    """

    cookies = None
    verbose = False
    jsonReader = None
    contentType = None
    handleErrors = True

    def __init__(self, app, contentType="application/json-rpc", jsonReader=None,
        verbose=0, username=None, password=None, bearer=None):
        super(JSONRPCTestTransport, self).__init__(contentType=contentType,
            jsonReader=jsonReader, verbose=verbose)
        self.app = app
        # optional crednetials
        self.username = username
        self.password = password
        self.bearer = bearer
        # store cookies between consecutive requests
        self.cookies = SimpleCookie()

    # cookies
    def httpCookie(self, path):
         """Return self.cookies as an HTTP_COOKIE environment value."""
         l = [m.OutputString().split(';')[0] for m in self.cookies.values()
              if path.startswith(m['path'])]
         return '; '.join(l)

    def loadCookies(self, envstring):
        self.cookies.load(envstring)

    def saveCookies(self, response):
        """Save cookies from the response."""
        for k,v in response._cookies.items():
            k = k.encode('utf8')
            self.cookies[k] = v['value'].encode('utf8')
            if v.has_key('path'):
                self.cookies[k]['path'] = v['path']

    def getEnviron(self, method, path, protocol):
        """Get environment based on first input line

        The first request line looks  somethng like:

        POST / HTTP/1.0\n

        """
        if SCHEME_RE.search(path):
            scheme, netloc, path, qs, fragment = urlparse.urlsplit(path)
            if fragment:
                raise TypeError(
                    "Path cannot contain a fragment (%r)" % fragment)
            if qs:
                path += '?' + qs
            if ':' not in netloc:
                if scheme == 'http':
                    netloc += ':80'
                elif scheme == 'https':
                    netloc += ':443'
                else:
                    raise TypeError("Unknown scheme: %r" % scheme)
        else:
            scheme = 'http'
            netloc = 'localhost:90'
        if path and '?' in path:
            path_info, query_string = path.split('?', 1)
            path_info = url_unquote(path_info)
        else:
            path_info = url_unquote(path)
            query_string = ''
        return {
            'HTTP_COOKIE': self.httpCookie(path),
            'HTTP_HOST': netloc,
            'HTTP_REFERER': netloc,
            'REQUEST_METHOD': method,
            'SCRIPT_NAME': '',
            'PATH_INFO': path_info or '',
            'QUERY_STRING': query_string,
            'SERVER_NAME': netloc.split(':')[0],
            'SERVER_PORT': netloc.split(':')[1],
            'SERVER_PROTOCOL': protocol,
        }

    # request handling
    def request(self, host, handler, request_body, verbose=0):
        """Handle request"""
        if not handler:
            handler = '/'
        request = "POST %s HTTP/1.0\n" % (handler,)
        request += "Content-Length: %i\n" % len(request_body)
        request += "Content-Type: application/json-rpc\n"

        host, extra_headers, x509 = self.get_host_info(host)
        if extra_headers:
            request += "Authorization: %s\n" % (
                dict(extra_headers)["Authorization"])
        elif self.username is not None and self.password is not None:
            credentials = "%s:%s" % (self.username, self.password)
            request += "Authorization: Basic %s\n" % base64.encodestring(
                credentials).replace("\012", "")
        elif self.bearer is not None:
            request += "Authorization: Bearer %s\n" % self.bearer

        request += "\n" + request_body
        response = self.doRequest(request, handleErrors=self.handleErrors)
        errcode = response.getStatus()
        errmsg = response.getStatusString()
        # This is not the same way that the normal transport deals with the
        # headers.
        headers = response.getHeaders()

        if errcode != 200:
            raise p01.json.exceptions.ProtocolError(host + handler, errcode,
                errmsg, headers)

        return self._parse_response(StringIO(response.getBody()), sock=None)

    def doPublish(self, instream, environment):
        """Publish input stream and return request"""
        request = self.app.publisher(instream, environment)
        return self.app.publisher.publish(request,
            handleErrors=self.handleErrors)

    def doRequest(self, request_string, handleErrors=True):
        """Process request and return response"""
        # commit work done by previous request
        transaction.commit()

        # discard leading white space to make call layout simpler
        request_string = request_string.lstrip()

        # split off and parse the command line
        l = request_string.find('\n')
        command_line = request_string[:l].rstrip()
        method, path, protocol = command_line.split()
        environment = self.getEnviron(method, path, protocol)

        request_string = request_string[l+1:]
        instream = StringIO(request_string)

        headers = [split_header(header)
                   for header in rfc822.Message(instream).headers]
        for name, value in headers:
            name = ('_'.join(name.upper().split('-')))
            if name not in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
                name = 'HTTP_' + name
            environment[name] = value.rstrip()

        auth_key = 'HTTP_AUTHORIZATION'
        if environment.has_key(auth_key):
            environment[auth_key] = auth_header(environment[auth_key])

        old_site = hooks.getSite()
        hooks.setSite(None)

        request = self.doPublish(instream, environment)

        response = ResponseWrapper(
            request.response, path,
            omit=('x-content-type-warning', 'x-powered-by'),
            )

        self.saveCookies(response)
        hooks.setSite(old_site)

        return response


def getJSONRPCTestProxy(uri, app=None, transport=None, encoding=None,
    verbose=None, jsonId=None, handleErrors=True, jsonVersion=JSON_RPC_VERSION,
    contentType="application/json-rpc", jsonReader=None, jsonWriter=None,
    username=None, password=None, bearer=None, cookies=None):
    """Test JSONRPCProxy using wsgi app and it's publisher for processing"""
    if transport is not None and username is not None and password is not None:
        raise Exception(
            "getJSONRPCTestProxy can't use transport and username and password. "
            "Use only transport or username and password. Setup authentication "
            "in your own transport class if needed.")
    if app is None:
        app = zope.component.getUtility(p01.publisher.interfaces.IApplication)
    publisher = p01.publisher.publisher.Publisher
    wsgi_app = p01.publisher.application.WSGIApplication(app, publisher,
        handleErrors)
    if transport is None:
        transport = JSONRPCTestTransport(wsgi_app, contentType=contentType,
            jsonReader=jsonReader, verbose=verbose, username=username,
            password=password, bearer=bearer)
    if isinstance(transport, JSONRPCTestTransport):
        transport.handleErrors = handleErrors
    if cookies is not None:
        transport.cookies = cookies
    return p01.json.proxy.JSONRPCProxy(uri, transport=transport,
        encoding=encoding, jsonId=jsonId, jsonVersion=jsonVersion,
        contentType=contentType, jsonWriter=jsonWriter, verbose=verbose)


def getCookies(proxy):
    """Get response cookies from proxy"""
    return proxy._JSONRPCProxy__transport.cookies

def setCookies(proxy, cookies):
    """Set request cookies given from previous proxy to new proxy"""
    proxy._JSONRPCProxy__transport.cookies = cookies


###############################################################################
#
# Test layer
#
###############################################################################

p01.publisher.testing.defineLayer("JSONRPCTestingLayer", "ftesting.zcml",
    allow_teardown=True)


###############################################################################
#
# Test component
#
###############################################################################

class IJSONRPCTestLayer(layer.IJSONRPCLayer):
    """JSON-RPC test layer interface used for zcml testing."""


class IJSONRPCTestSkin(IJSONRPCTestLayer):
    """The IJSONRPCTestSkin testing skin based on IJSONRPCLayer."""


class IA(zope.interface.Interface):
    """First content stub interface."""


class A(persistent.Persistent):
    """First content stub."""

    zope.interface.implements(IA)


class IB(zope.interface.Interface):
    """First content stub interface."""


class B(persistent.Persistent):
    """First content stub."""

    zope.interface.implements(IB)


class MethodsA(MethodPublisher):
    """Method publisher test class."""

    def hello(self):
        return "Hello A World"


class MethodsB(MethodPublisher):
    """Method publisher test class."""

    def hello(self):
        return "Hello B World"


class TestRequest(JSONRPCRequest):
    """modeled after zope.publisher.xmlrpc.TestRequest"""
    def __init__(self, body_instream=None, environ=None,
                 response=None, **kw):

        _testEnv =  {
            'SERVER_URL':         'http://127.0.0.1',
            'HTTP_HOST':          '127.0.0.1',
            'CONTENT_LENGTH':     '0',
            'GATEWAY_INTERFACE':  'TestFooInterface/1.0',
            }

        if environ:
            _testEnv.update(environ)
        if kw:
            _testEnv.update(kw)
        if body_instream is None:
            body_instream = StringIO('')

        super(TestRequest, self).__init__(
            body_instream, _testEnv, response)


class CookieHandler(object):

    def __init__(self, *args, **kw):
        # Somewhere to store cookies between consecutive requests
        self.cookies = SimpleCookie()
        super(CookieHandler, self).__init__(*args, **kw)

    def httpCookie(self, path):
        """Return self.cookies as an HTTP_COOKIE environment value."""
        l = [m.OutputString().split(';')[0] for m in self.cookies.values()
             if path.startswith(m['path'])]
        return '; '.join(l)

    def loadCookies(self, envstring):
        self.cookies.load(envstring)

    def saveCookies(self, response):
        """Save cookies from the response."""
        # Urgh - need to play with the response's privates to extract
        # cookies that have been set
        # TODO: extend the IHTTPRequest interface to allow access to all
        # cookies
        # TODO: handle cookie expirations
        for k, v in response._cookies.items():
            k = k.encode('utf8') if bytes is str else k
            val = v['value']
            val = val.encode('utf8') if bytes is str else val
            self.cookies[k] = val
            if 'path' in v:
                self.cookies[k]['path'] = v['path']


# class HTTPCaller(functional.HTTPCaller):
#     """HTTPCaller for JSON."""

#     def chooseRequestClass(self, method, path, environment):
#         """Choose and return a request class and a publication class"""
#         request_cls, publication_cls = \
#             super(HTTPCaller, self).chooseRequestClass(method, path,
#                 environment)

#         content_type = environment.get('CONTENT_TYPE', '')
#         is_json = content_type.startswith('application/json')

#         if method in ('GET', 'POST', 'HEAD'):
#             if (method == 'POST' and is_json):
#                 request_cls = JSONRPCRequest
#                 publication_cls = JSONRPCPublication

#         return request_cls, publication_cls


# class HTTPCaller(CookieHandler):
#     """HTTPCaller for JSON."""

#     def chooseRequestClass(self, method, path, environment):
#         """Choose and return a request class and a publication class"""
#         request_cls, publication_cls = \
#             super(HTTPCaller, self).chooseRequestClass(method, path,
#                 environment)

#         content_type = environment.get('CONTENT_TYPE', '')
#         is_json = content_type.startswith('application/json')

#         if method in ('GET', 'POST', 'HEAD'):
#             if (method == 'POST' and is_json):
#                 request_cls = JSONRPCRequest
#                 publication_cls = JSONRPCPublication

#         return request_cls, publication_cls

#     def __call__(self, request_string, handle_errors=True, form=None):
#         # Commit work done by previous python code.
#         commit()

#         # Discard leading white space to make call layout simpler
#         request_string = request_string.lstrip()

#         # split off and parse the command line
#         l = request_string.find('\n')
#         command_line = request_string[:l].rstrip()
#         request_string = request_string[l + 1:]
#         method, path, protocol = command_line.split()

#         # If we don't feed bytes to Python 3, it gets stuck in a loop
#         # and ultimately raises HTTPException: got more than 100 headers.
#         instream = io.BytesIO(request_string.encode("latin-1")
#                               if not isinstance(request_string, bytes)
#                               else request_string)
#         environment = {
#             "HTTP_COOKIE": self.httpCookie(path),
#             "HTTP_HOST": 'localhost',
#             "REQUEST_METHOD": method,
#             "SERVER_PROTOCOL": protocol,
#             "REMOTE_ADDR": '127.0.0.1',
#         }

#         headers = [split_header(header)
#                    for header in headers_factory(instream).headers]
#         for name, value in headers:
#             name = ('_'.join(name.upper().split('-')))
#             if name not in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
#                 name = 'HTTP_' + name
#             environment[name] = value.rstrip()

#         auth_key = 'HTTP_AUTHORIZATION'
#         if auth_key in environment:
#             environment[auth_key] = auth_header(environment[auth_key])

#         old_site = hooks.getSite()
#         hooks.setSite(None)

#         request_cls, publication_cls = self.chooseRequestClass(method, path,
#                                                                environment)
#         app = FunctionalTestSetup().getApplication()

#         request = app._request(
#             path, instream,
#             environment=environment,
#             request=request_cls, publication=publication_cls)
#         if ISkinnable.providedBy(request):
#             # only ISkinnable requests have skins
#             setDefaultSkin(request)

#         if form is not None:
#             if request.form:
#                 raise ValueError("only one set of form values can be provided")
#             request.form = form

#         # request = publish(request, handle_errors=handle_errors)
#         request = self.app.publisher.publish(request,
#             handle_errors=handle_errors)

#         response = ResponseWrapper(
#             request.response, path, request,
#             omit=('x-content-type-warning', 'x-powered-by'),
#             )

#         self.saveCookies(response)
#         hooks.setSite(old_site)

#         # sync Python connection:
#         getRootFolder()._p_jar.sync()

#         return response


###############################################################################
#
# Doctest setup
#
###############################################################################

def FunctionalDocFileSuite(*paths, **kw):
    # use our custom HTTPCaller and layer
    globs = kw.setdefault('globs', {})
    globs['getRootFolder'] = getRootFolder
    kw['package'] = doctest._normalize_module(kw.get('package'))
    # _prepare_doctest_keywords(kw)
    p01.publisher.testing.prepareDocTestKeywords(kw)
    suite = doctest.DocFileSuite(*paths, **kw)
    suite.layer = JSONRPCTestingLayer
    return suite


###############################################################################
#
# Test helper, make us independent from zope.app.testing
#
###############################################################################

# Setup of test text files as modules
import sys

# Evil hack to make pickling work with classes defined in doc tests
class NoCopyDict(dict):
    def copy(self):
        return self

class FakeModule:
    """A fake module."""

    def __init__(self, dict):
        self.__dict = dict

    def __getattr__(self, name):
        try:
            return self.__dict[name]
        except KeyError:
            raise AttributeError(name)


def setUpTestAsModule(test, name=None):
    if name is None:
        if test.globs.haskey('__name__'):
            name = test.globs['__name__']
        else:
            name = test.globs.name

    test.globs['__name__'] = name
    test.globs = NoCopyDict(test.globs)
    sys.modules[name] = FakeModule(test.globs)


def tearDownTestAsModule(test):
    del sys.modules[test.globs['__name__']]
    test.globs.clear()


###############################################################################
#
# Unittest setup
#
###############################################################################

def setUp(test):
    setUpTestAsModule(test, name='README')


def tearDown(test):
    # ensure that we cleanup everything
    zope.testing.cleanup.cleanUp()
    tearDownTestAsModule(test)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""

import zope.component
import zope.traversing.namespace
from zope.location.interfaces import LocationError
from zope.publisher.skinnable import applySkin
from zope.interface.interfaces import ComponentLookupError

from p01.jsonrpc import interfaces


class skin(zope.traversing.namespace.view):
    """JSONRPC skin namespace"""

    def traverse(self, name, ignored):
        self.request.shiftNameToApplication()
        try:
            skin = zope.component.getUtility(interfaces.IJSONRPCSkinType, name)
        except ComponentLookupError:
            raise LocationError("++skin++%s" % name)
        applySkin(self.request, skin)
        return self.context



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""

import zope.interface
from p01.jsonrpc import interfaces


class JSONRPCException(Exception):
    """Base class for JSON-RPC exception."""

    zope.interface.implements(interfaces.IJSONRPCException)


class ParseError(JSONRPCException):
    """Invalid JSON. An error occurred on the server while parsing the JSON text."""


class InvalidRequest(JSONRPCException):
    """The received JSON not a valid JSON-RPC Request. """


class MethodNotFound(JSONRPCException):
    """The requested remote-procedure does not exist, is not available."""


class InvalidParams(JSONRPCException):
    """Invalid method parameters."""


class InternalError(JSONRPCException):
    """Internal JSON-RPC error."""



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import zope.schema
import zope.interface
from zope.publisher.interfaces import IPublication
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces import ISkinnable
from zope.publisher.interfaces import ISkinType
from zope.publisher.interfaces.http import IHTTPApplicationRequest
from zope.publisher.interfaces.http import IHTTPCredentials
from zope.publisher.interfaces.http import IHTTPRequest
from zope.app.publication.interfaces import IRequestFactory

JSON_CHARSETS = ('utf-8','utf-16', 'utf-32')


class IMethodPublisher(zope.interface.Interface):
    """Marker interface for an object that wants to publish methods."""


class IJSONRPCRequestFactory(IRequestFactory):
    """Browser request factory"""


class IJSONRPCPublisher(IPublishTraverse):
    """JSON-RPC Publisher
    like zope.publisher.interfaces.xmlrpc.IXMLRPCPublisher
    """


class IJSONRPCPublication(IPublication):
    """Publication for JOSN-RPC-based protocol."""


class IJSONRPCSkinType(ISkinType):
    """A skin is a set of layers."""


class IJSONRPCApplicationRequest(IHTTPApplicationRequest):
    """HTTP application request."""


class IJSONRPCRequest(IJSONRPCApplicationRequest, IHTTPCredentials,
    IHTTPRequest, ISkinnable):
    """JSON-RPC request."""

    jsonID = zope.interface.Attribute("""JSON-RPC ID for the request""")


class IJSONRPCException(zope.interface.Interface):
    """JSON-RPC error"""


class IJSONRPCErrorView(zope.interface.Interface):
    """Error view base class used by ZopePublications error handling.
    """

    code = zope.schema.Int(
        title=u'Error code',
        description=u'JSON-RPC error code',
        default=-32603,
        required=True)

    message = zope.schema.Text(
        title=u'Error message',
        description=u'JSON-RPC error message',
        default=u'Internal error',
        required=True)

    data = zope.schema.Text(
        title=u'Error data',
        description=u'JSON-RPC error data',
        default=u'',
        required=True)

    def __init__(self):
        """Adapts an error and a request."""

    def __call__(self):
        """Must return itself by calling."""



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import zope.interface
import zope.component
import zope.configuration.fields
import zope.interface
import zope.schema
import zope.security.zcml
from zope.publisher.interfaces import IDefaultSkin

from zope.interface import Interface
from zope.security.checker import CheckerPublic, Checker
from zope.configuration.exceptions import ConfigurationError

from zope.component.interface import provideInterface
from zope.component.zcml import handler

from p01.jsonrpc import interfaces
from p01.jsonrpc.publisher import MethodPublisher


class IJSONRPCDirective(zope.interface.Interface):
    """Directive for JSONRPC methods."""

    for_ = zope.configuration.fields.GlobalObject(
        title=u"Published Object Type",
        description=u"""The types of objects to be published via JSONRPC

        This can be expressed with either a class or an interface
        """,
        required=True,
        )

    interface = zope.configuration.fields.Tokens(
        title=u"Interface to be published.",
        required=False,
        value_type=zope.configuration.fields.GlobalInterface()
        )

    methods = zope.configuration.fields.Tokens(
        title=u"Methods (or attributes) to be published",
        required=False,
        value_type=zope.configuration.fields.PythonIdentifier()
        )

    class_ = zope.configuration.fields.GlobalObject(
        title=u"Class",
        description=u"A class that provides attributes used by the view.",
        required=False
        )

    permission = zope.security.zcml.Permission(
        title=u"Permission",
        description=u"""The permission needed to use the view.

        If this option is used and a name is given for the view, then
        the names defined by the given methods or interfaces will be
        under the given permission.

        If a name is not given for the view, then, this option is required and
        the given permission is required to call the individual views defined
        by the given interface and methods.

        (See the name attribute.)

        If no permission is given, then permissions should be declared
        for the view using other means, such as the class directive.
        """,
        required=False)

    name = zope.schema.TextLine(
        title=u"The name of the view.",
        description=u"""

        If a name is given, then rpc methods are accessed by
        traversing the name and then accessing the methods.  In this
        case, the class should implement
        zope.pubisher.interfaces.IPublishTraverse.

        If no name is provided, then the names given by the attributes
        and interfaces are published directly as callable views.

        """,
        required=False,
        )

    layer = zope.configuration.fields.GlobalInterface(
        title=u"The layer the view is declared for",
        description=u"The default layer for which the default view is "
                    u"applicable. By default it is applied to all layers.",
        default=interfaces.IJSONRPCRequest,
        required=False
        )


def jsonrpc(_context, for_=None, interface=None, methods=None, class_=None,
    permission=None, name=None, layer=None):

    interface = interface or []
    methods = methods or []

    if layer is not None:
        if not layer.extends(interfaces.IJSONRPCRequest):
            raise ConfigurationError(
                "The layer interface must extend `IJSONRPCRequest`.")
    else:
        layer = interfaces.IJSONRPCRequest

    # If there were special permission settings provided, then use them
    if permission == 'zope.Public':
        permission = CheckerPublic

    require = {}
    for attr_name in methods:
        require[attr_name] = permission

    if interface:
        for iface in interface:
            for field_name in iface:
                require[field_name] = permission
            _context.action(
                discriminator = None,
                callable = provideInterface,
                args = ('', for_)
                )

    # Make sure that the class inherits MethodPublisher, so that the views
    # have a location
    if class_ is None:
        class_ = original_class = MethodPublisher
    else:
        original_class = class_
        class_ = type(class_.__name__, (class_, MethodPublisher), {})

    if name:
        # Register a single jsonrpc view

        if permission:
            checker = Checker(require)

            def proxyView(context, request, class_=class_, checker=checker):
                view = class_(context, request)
                # We need this in case the resource gets unwrapped and
                # needs to be rewrapped
                view.__Security_checker__ = checker
                return view

            class_ = proxyView
            class_.factory = original_class
        else:
            # No permission was defined, so we defer to the checker
            # of the original class
            def proxyView(context, request, class_=class_):
                view = class_(context, request)
                view.__Security_checker__ = getCheckerForInstancesOf(
                    original_class)
                return view
            class_ = proxyView
            class_.factory = original_class

        # Register the new view.
        _context.action(
            discriminator = ('jsonrpc', for_, name, layer),
            callable = handler,
            args = ('registerAdapter',
                    class_, (for_, layer), Interface, name,
                    _context.info)
            )
    else:
        if permission:
            checker = Checker({'__call__': permission})
        else:
            raise ConfigurationError(
              "JSONRPC view has neither a name nor a permission. "
              "You have to specify at least one of the two.")

        for name in require:
            # create a new callable class with a security checker;
            cdict = {'__Security_checker__': checker,
                     '__call__': getattr(class_, name)}
            new_class = type(class_.__name__, (class_,), cdict)
            _context.action(
                discriminator = ('jsonrpc', for_, name, layer),
                callable = handler,
                args = ('registerAdapter',
                        new_class, (for_, layer), Interface, name,
                        _context.info)
                )

    # Register the used interfaces with the site manager
    if for_ is not None:
        _context.action(
            discriminator = None,
            callable = provideInterface,
            args = ('', for_)
            )


class IDefaultJSONRPCSkinDirective(zope.interface.Interface):
    """Sets the default JSON-RPC skin."""

    name = zope.schema.TextLine(
        title=u"Default skin name",
        description=u"Default skin name",
        required=True
        )


def setDefaultJSONRPCSkin(name, info=''):
    """Set the default skin."""
    skin = zope.component.getUtility(interfaces.IJSONRPCSkinType, name=name)
    handler('registerAdapter', skin, (interfaces.IJSONRPCRequest,),
        IDefaultSkin, '', info),


def defaultJSONRPCSkin(_context, name):

    _context.action(
        discriminator = 'setDefaultJSONRPCSkin',
        callable = setDefaultJSONRPCSkin,
        args = (name, _context.info)
        )
